package ru.dbotthepony.kstarbound

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.internal.bind.JsonTreeReader
import com.google.gson.stream.JsonReader
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.defs.Json2Function
import ru.dbotthepony.kstarbound.defs.JsonFunction
import ru.dbotthepony.kstarbound.defs.Species
import ru.dbotthepony.kstarbound.defs.StatusEffectDefinition
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.BackArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ChestArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.CurrencyItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.FlashlightDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.HarvestingToolPrototype
import ru.dbotthepony.kstarbound.defs.item.impl.HeadArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.ItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LegsArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.LiquidItemDefinition
import ru.dbotthepony.kstarbound.defs.item.impl.MaterialItemDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterSkillDefinition
import ru.dbotthepony.kstarbound.defs.monster.MonsterTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.NpcTypeDefinition
import ru.dbotthepony.kstarbound.defs.npc.TenantDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.particle.ParticleDefinition
import ru.dbotthepony.kstarbound.defs.player.TechDefinition
import ru.dbotthepony.kstarbound.defs.projectile.ProjectileDefinition
import ru.dbotthepony.kstarbound.defs.quest.QuestTemplate
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.util.AssetPathStack
import java.util.concurrent.ExecutorService
import java.util.concurrent.ForkJoinPool
import java.util.concurrent.ForkJoinTask
import java.util.concurrent.Future

object Registries {
	private val LOGGER = LogManager.getLogger()

	private val registries = ArrayList<Registry<*>>()

	val tiles = Registry<TileDefinition>("tiles").also(registries::add)
	val tileModifiers = Registry<MaterialModifier>("tile modifiers").also(registries::add)
	val liquid = Registry<LiquidDefinition>("liquid").also(registries::add)
	val species = Registry<Species>("species").also(registries::add)
	val statusEffects = Registry<StatusEffectDefinition>("status effects").also(registries::add)
	val particles = Registry<ParticleDefinition>("particles").also(registries::add)
	val items = Registry<IItemDefinition>("items").also(registries::add)
	val questTemplates = Registry<QuestTemplate>("quest templates").also(registries::add)
	val techs = Registry<TechDefinition>("techs").also(registries::add)
	val jsonFunctions = Registry<JsonFunction>("json functions").also(registries::add)
	val json2Functions = Registry<Json2Function>("json 2functions").also(registries::add)
	val npcTypes = Registry<NpcTypeDefinition>("npc types").also(registries::add)
	val projectiles = Registry<ProjectileDefinition>("projectiles").also(registries::add)
	val tenants = Registry<TenantDefinition>("tenants").also(registries::add)
	val treasurePools = Registry<TreasurePoolDefinition>("treasure pools").also(registries::add)
	val monsterSkills = Registry<MonsterSkillDefinition>("monster skills").also(registries::add)
	val monsterTypes = Registry<MonsterTypeDefinition>("monster types").also(registries::add)
	val worldObjects = Registry<ObjectDefinition>("objects").also(registries::add)

	private fun <T> key(mapper: (T) -> String): (T) -> Pair<String, Int?> {
		return { mapper.invoke(it) to null }
	}

	private fun <T> key(mapper: (T) -> String, mapperInt: (T) -> Int): (T) -> Pair<String, Int> {
		return { mapper.invoke(it) to mapperInt.invoke(it) }
	}

	fun validate(): Boolean {
		var any = false

		any = !tiles.validate() || any
		any = !tileModifiers.validate() || any
		any = !liquid.validate() || any
		any = !species.validate() || any
		any = !statusEffects.validate() || any
		any = !particles.validate() || any
		any = !items.validate() || any
		any = !questTemplates.validate() || any
		any = !techs.validate() || any
		any = !jsonFunctions.validate() || any
		any = !json2Functions.validate() || any
		any = !npcTypes.validate() || any
		any = !projectiles.validate() || any
		any = !tenants.validate() || any
		any = !treasurePools.validate() || any
		any = !monsterSkills.validate() || any
		any = !monsterTypes.validate() || any
		any = !worldObjects.validate() || any

		return !any
	}

	private inline fun <reified T : Any> loadRegistry(
		executor: ExecutorService,
		registry: Registry<T>,
		files: List<IStarboundFile>,
		noinline keyProvider: (T) -> Pair<String, Int?>
	): List<Future<*>> {
		val adapter by lazy { Starbound.gson.getAdapter(T::class.java) }
		val elementAdapter by lazy { Starbound.gson.getAdapter(JsonElement::class.java) }

		return files.map { listedFile ->
			executor.submit {
				try {
					AssetPathStack(listedFile.computeDirectory()) {
						val elem = elementAdapter.read(listedFile.jsonReader())
						val read = adapter.fromJsonTree(elem)
						val keys = keyProvider(read)

						registry.add {
							if (keys.second != null)
								registry.add(keys.first, keys.second!!, read, elem, listedFile)
							else
								registry.add(keys.first, read, elem, listedFile)
						}
					}
				} catch (err: Throwable) {
					LOGGER.error("Loading ${registry.name} definition file $listedFile", err)
				}
			}
		}
	}

	fun finishLoad() {
		registries.forEach { it.finishLoad() }
	}

	fun load(fileTree: Map<String, List<IStarboundFile>>, executor: ExecutorService): List<Future<*>> {
		val tasks = ArrayList<Future<*>>()

		tasks.addAll(loadItemDefinitions(fileTree, executor))

		tasks.addAll(loadJsonFunctions(fileTree["functions"] ?: listOf(), executor))
		tasks.addAll(loadJson2Functions(fileTree["2functions"] ?: listOf(), executor))
		tasks.addAll(loadTreasurePools(fileTree["treasurepools"] ?: listOf(), executor))

		tasks.addAll(loadRegistry(executor, tiles, fileTree["material"] ?: listOf(), key(TileDefinition::materialName, TileDefinition::materialId)))
		tasks.addAll(loadRegistry(executor, tileModifiers, fileTree["matmod"] ?: listOf(), key(MaterialModifier::modName, MaterialModifier::modId)))
		tasks.addAll(loadRegistry(executor, liquid, fileTree["liquid"] ?: listOf(), key(LiquidDefinition::name, LiquidDefinition::liquidId)))

		tasks.addAll(loadRegistry(executor, worldObjects, fileTree["object"] ?: listOf(), key(ObjectDefinition::objectName)))
		tasks.addAll(loadRegistry(executor, statusEffects, fileTree["statuseffect"] ?: listOf(), key(StatusEffectDefinition::name)))
		tasks.addAll(loadRegistry(executor, species, fileTree["species"] ?: listOf(), key(Species::kind)))
		tasks.addAll(loadRegistry(executor, particles, fileTree["particle"] ?: listOf(), key(ParticleDefinition::kind)))
		tasks.addAll(loadRegistry(executor, questTemplates, fileTree["questtemplate"] ?: listOf(), key(QuestTemplate::id)))
		tasks.addAll(loadRegistry(executor, techs, fileTree["tech"] ?: listOf(), key(TechDefinition::name)))
		tasks.addAll(loadRegistry(executor, npcTypes, fileTree["npctype"] ?: listOf(), key(NpcTypeDefinition::type)))
		tasks.addAll(loadRegistry(executor, monsterSkills, fileTree["monsterskill"] ?: listOf(), key(MonsterSkillDefinition::name)))

		return tasks
	}

	private fun loadItemDefinitions(files: Map<String, Collection<IStarboundFile>>, executor: ExecutorService): List<Future<*>> {
		val fileMap = mapOf(
			"item" to ItemDefinition::class.java,
			"currency" to CurrencyItemDefinition::class.java,
			"liqitem" to LiquidItemDefinition::class.java,
			"matitem" to MaterialItemDefinition::class.java,
			"flashlight" to FlashlightDefinition::class.java,
			"harvestingtool" to HarvestingToolPrototype::class.java,
			"head" to HeadArmorItemDefinition::class.java,
			"chest" to ChestArmorItemDefinition::class.java,
			"legs" to LegsArmorItemDefinition::class.java,
			"back" to BackArmorItemDefinition::class.java,
		)

		val tasks = ArrayList<Future<*>>()
		val objects = Starbound.gson.getAdapter(JsonObject::class.java)

		for ((ext, clazz) in fileMap) {
			val fileList = files[ext] ?: continue
			val adapter by lazy { Starbound.gson.getAdapter(clazz) }

			for (listedFile in fileList) {
				tasks.add(executor.submit {
					try {
						val json = objects.read(listedFile.jsonReader())
						val def = AssetPathStack(listedFile.computeDirectory()) { adapter.fromJsonTree(json) }

						items.add {
							items.add(def.itemName, def, json, listedFile)
						}
					} catch (err: Throwable) {
						LOGGER.error("Loading item definition file $listedFile", err)
					}
				})
			}
		}

		return tasks
	}

	private fun loadJsonFunctions(files: Collection<IStarboundFile>, executor: ExecutorService): List<Future<*>> {
		return files.map { listedFile ->
			executor.submit {
				try {
					val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

					for ((k, v) in json.entrySet()) {
						try {
							val fn = Starbound.gson.fromJson<JsonFunction>(JsonTreeReader(v), JsonFunction::class.java)
							jsonFunctions.add(k, fn, v, listedFile)
						} catch (err: Exception) {
							LOGGER.error("Loading json function definition $k from file $listedFile", err)
						}
					}
				} catch (err: Exception) {
					LOGGER.error("Loading json function definition $listedFile", err)
				}
			}
		}
	}

	private fun loadJson2Functions(files: Collection<IStarboundFile>, executor: ExecutorService): List<Future<*>> {
		return files.map { listedFile ->
			executor.submit {
				try {
					val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

					for ((k, v) in json.entrySet()) {
						try {
							val fn = Starbound.gson.fromJson<Json2Function>(JsonTreeReader(v), Json2Function::class.java)
							json2Functions.add(k, fn, v, listedFile)
						} catch (err: Throwable) {
							LOGGER.error("Loading json 2function definition $k from file $listedFile", err)
						}
					}
				} catch (err: Exception) {
					LOGGER.error("Loading json 2function definition $listedFile", err)
				}
			}
		}
	}

	private fun loadTreasurePools(files: Collection<IStarboundFile>, executor: ExecutorService): List<Future<*>> {
		return files.map { listedFile ->
			executor.submit {
				try {
					val json = Starbound.gson.getAdapter(JsonObject::class.java).read(JsonReader(listedFile.reader()).also { it.isLenient = true })

					for ((k, v) in json.entrySet()) {
						try {
							val result = Starbound.gson.fromJson<TreasurePoolDefinition>(JsonTreeReader(v), TreasurePoolDefinition::class.java)
							result.name = k
							treasurePools.add(result.name, result, v, listedFile)
						} catch (err: Throwable) {
							LOGGER.error("Loading treasure pool definition $k from file $listedFile", err)
						}
					}
				} catch (err: Exception) {
					LOGGER.error("Loading treasure pool definition $listedFile", err)
				}
			}
		}
	}
}
